def producer_to_projector(key, producer):
    def projector(instance):
        return {key: producer(instance)}

    return projector


def combine(*projectors):
    """
    Given a list of projectors as *args, return another projector which calls each
    projector in turn and merges the resulting dictionaries.
    """

    def combined(instance):
        result = {}
        for projector in projectors:
            projection = projector(instance)
            if not isinstance(projection, dict):
                raise TypeError(f"Projector {projector} did not return a dictionary")
            result.update(projection)
        return result

    return combined


def noop(instance):
    return {}



from django.db.models import Count, Sum
from django_readers import producers, projectors, qs
from operator import itemgetter


def producer_to_projector(name, pair):
    prepare, produce = pair
    return prepare, projectors.producer_to_projector(name, produce)


def field(name, *, transform_value=None, transform_value_if_none=False):
    return qs.include_fields(name), producers.attr(
        name,
        transform_value=transform_value,
        transform_value_if_none=transform_value_if_none,
    )


def combine(*pairs):
    """
    Given a list of pairs as *args, return a pair which pipes together the prepare
    function from each pair, and combines the project function from each pair.
    """
    prepare_fns, project_fns = zip(*pairs)
    return qs.pipe(*prepare_fns), projectors.combine(*project_fns)


discard_projector = itemgetter(0)


discard_queryset_function = itemgetter(1)


def field_display(name):
    """
    Works with Django's get_FOO_display mechanism for fields with choices set. Given
    the name of a field, returns a producer that calls get_<name>_display.
    """
    return qs.include_fields(name), producers.method(f"get_{name}_display")


def annotate(*args, **kwargs):
    """
    Return a pair that adds an annotation to the queryset and produces the value. Like
    the annotate method on QuerySet, this can take either a positional argument or a
    keyword argument. Unlike the annotate method, this can only handle a single
    annotation at a time. Also, this function can optionally take transform_value and
    tranform_value_if_none arguments, which are passed to the producer.
    """
    transform_value = kwargs.pop("transform_value", None)
    transform_value_if_none = kwargs.pop("transform_value_if_none", False)

    if len(args) + len(kwargs) != 1:
        raise ValueError("Only a single annotation is supported")

    annotations = kwargs or {args[0].default_alias: args[0]}
    name, annotation = next(iter(annotations.items()))

    return (
        qs.annotate(**{name: annotation}),
        producers.attr(
            name,
            transform_value=transform_value,
            transform_value_if_none=transform_value_if_none,
        ),
    )


def count(name, *args, **kwargs):
    return annotate(
        **{f"{name}_count": Count(name, *args, **kwargs)},
    )


def has(name, *args, **kwargs):
    return annotate(
        **{f"{name}_count": Count(name, *args, **kwargs)}, transform_value=bool
    )


def sum(name, *args, **kwargs):
    return annotate(
        **{f"{name}_sum": Sum(name, *args, **kwargs)},
    )


def filter(*args, **kwargs):
    return qs.filter(*args, **kwargs), projectors.noop


def exclude(*args, **kwargs):
    return qs.exclude(*args, **kwargs), projectors.noop


def order_by(*args, **kwargs):
    return qs.order_by(*args, **kwargs), projectors.noop


"""
Below are pair functions which return the various queryset functions that prefetch
relationships of various types, and then produce those related objects.

There are functions for forward, reverse or many-to-many relationships, and then
a `relationship` function which selects the correct one by introspecting the
model. It shouldn't usually be necessary to use the manual functions unless you're
doing something weird, like providing a custom queryset.
"""


def forward_relationship(name, related_queryset, relationship_pair, to_attr=None):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.prefetch_forward_relationship(
        name, related_queryset, prepare_related_queryset, to_attr
    )
    return prepare, producers.relationship(to_attr or name, project_relationship)


def reverse_relationship(
    name, related_name, related_queryset, relationship_pair, to_attr=None
):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.prefetch_reverse_relationship(
        name, related_name, related_queryset, prepare_related_queryset, to_attr
    )
    return prepare, producers.relationship(to_attr or name, project_relationship)


def many_to_many_relationship(name, related_queryset, relationship_pair, to_attr=None):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.prefetch_many_to_many_relationship(
        name, related_queryset, prepare_related_queryset, to_attr
    )
    return prepare, producers.relationship(to_attr or name, project_relationship)


def relationship(name, relationship_pair, to_attr=None):
    prepare_related_queryset, project_relationship = relationship_pair
    prepare = qs.auto_prefetch_relationship(name, prepare_related_queryset, to_attr)
    return prepare, producers.relationship(to_attr or name, project_relationship)


def pk_list(name, to_attr=None):
    return (
        qs.auto_prefetch_relationship(name, qs.include_fields("pk"), to_attr=to_attr),
        producers.pk_list(to_attr or name),
    )



from django_readers import pairs
from django_readers.utils import queries_disabled


def process_item(item):
    if isinstance(item, str):
        item = {item: item}
    if isinstance(item, dict):
        return pairs.combine(
            *[
                relationship_or_wrap(name, child_spec)
                for name, child_spec in item.items()
            ]
        )
    return item


def process(spec):
    return queries_disabled(pairs.combine(*(process_item(item) for item in spec)))


def relationship(name, relationship_spec, to_attr=None):
    return pairs.producer_to_projector(
        to_attr or name, pairs.relationship(name, process(relationship_spec), to_attr)
    )


def relationship_or_wrap(name, child_spec):
    if isinstance(child_spec, str):
        producer_pair = pairs.field(child_spec)
    elif isinstance(child_spec, list):
        producer_pair = pairs.relationship(name, process(child_spec))
    elif isinstance(child_spec, dict):
        if len(child_spec) != 1:
            raise ValueError("Aliased relationship spec must contain only one key")
        relationship_name, relationship_spec = next(iter(child_spec.items()))
        producer_pair = pairs.relationship(
            relationship_name, process(relationship_spec)
        )
    else:
        producer_pair = child_spec
    return pairs.producer_to_projector(name, producer_pair)



from django.core.exceptions import ImproperlyConfigured
from django.utils.functional import cached_property
from django_readers import specs


class ProjectionSerializer:
    def __init__(self, data=None, many=False, context=None):
        self.many = many
        self._data = data
        self.context = context

    @property
    def data(self):
        project = self.context["view"].project
        if self.many:
            return [project(item) for item in self._data]
        return project(self._data)


class SpecMixin:
    spec = None

    def get_spec(self):
        if self.spec is None:
            raise ImproperlyConfigured("SpecMixin requires spec or get_spec")
        return self.spec

    def get_reader_pair(self):
        return specs.process(self.get_spec())

    @cached_property
    def reader_pair(self):
        return self.get_reader_pair()

    @property
    def prepare(self):
        return self.reader_pair[0]

    @property
    def project(self):
        return self.reader_pair[1]

    def get_queryset(self):
        queryset = super().get_queryset()
        return self.prepare(queryset)

    def get_serializer_class(self):
        return ProjectionSerializer



from django.db.models import Prefetch, QuerySet
from django.db.models.constants import LOOKUP_SEP
from django.db.models.fields.related_descriptors import (
    ForwardManyToOneDescriptor,
    ForwardOneToOneDescriptor,
    ManyToManyDescriptor,
    ReverseManyToOneDescriptor,
    ReverseOneToOneDescriptor,
)


def _method_to_function(method):
    """
    Return a function that when called with any arguments will return another
    function that takes a queryset and calls the given queryset method, passing
    the queryset as the `self` parameter and forwarding along the other arguments.
    """

    def make_queryset_function(*args, **kwargs):
        def queryset_function(queryset):
            return method(queryset, *args, **kwargs)

        return queryset_function

    return make_queryset_function


"""
The following functions mirror all the default methods on the base
`QuerySet` that return a new `QuerySet`.
"""
filter = _method_to_function(QuerySet.filter)
all = _method_to_function(QuerySet.all)
exclude = _method_to_function(QuerySet.exclude)
select_related = _method_to_function(QuerySet.select_related)
prefetch_related = _method_to_function(QuerySet.prefetch_related)
order_by = _method_to_function(QuerySet.order_by)
distinct = _method_to_function(QuerySet.distinct)
extra = _method_to_function(QuerySet.extra)
defer = _method_to_function(QuerySet.defer)
only = _method_to_function(QuerySet.only)
using = _method_to_function(QuerySet.using)


def annotate(*args, **kwargs):
    def queryset_function(queryset):
        return include_fields("pk")(queryset.annotate(*args, **kwargs))

    return queryset_function


noop = all()  # a queryset function that does nothing


def include_fields(*fields):
    """
    Returns a queryset function that extends an already-`.only()`d queryset
    with more fields.
    """

    def fields_included(queryset):
        return queryset.only(*queryset.query.deferred_loading[0], *fields)

    return fields_included


def pipe(*fns):
    """
    Given a list of queryset functions as *args, return a new queryset function
    that calls each function in turn, passing the return value of each as the
    argument to the next.
    """

    def piped(queryset):
        for fn in fns:
            queryset = fn(queryset)
        return queryset

    return piped


def select_related_fields(*fields):
    """
    Like select_related, but selects only specific fields from the related objects
    """
    return pipe(
        select_related(*{field.rpartition(LOOKUP_SEP)[0] for field in fields}),
        include_fields(*fields),
    )


"""
The below functions return functions that use `prefetch_related` to efficiently load
related objects. We use `prefetch_related` to load all relationship types because this
means our functions can be recursive - we can apply pairs to the related querysets, all
the way down the tree.

There are six types of relationship from the point of view of the "main" object:

  * Forward one-to-one - a OneToOneField on the main object
  * Reverse one-to-one - a OneToOneField on the related object
  * Forward many-to-one - a ForeignKey on the main object
  * Reverse many-to-one - a ForeignKey on the related object
  * Forward many-to-many - a ManyToManyField on the main object
  * Reverse many-to-many - a ManyToManyField on the related object

ManyToManyFields are symmetrical, so the latter two collapse down to the same thing.
The forward one-to-one and many-to-one are identical as they both relate a single
related object to the main object. The reverse one-to-one and many-to-one are identical
except the former relates the main object to a single related object, and the latter
relates the main object to many related objects. Because the `projectors.relationship`
function already infers whether to iterate or project a single instance, we can collapse
these two functions into one as well.

There are functions for forward, reverse or many-to-many relationships, and then
an "auto" function which selects the correct relationship type by introspecting the
model. It shouldn't usually be necessary to use the manual functions unless you're
doing something weird, like providing a custom queryset.
"""


def prefetch_forward_relationship(
    name, related_queryset, prepare_related_queryset=noop, to_attr=None
):
    """
    Efficiently prefetch a forward relationship: one where the field on the "parent"
    queryset is a concrete field. We need to include this field in the query.
    """
    return pipe(
        include_fields(name),
        prefetch_related(
            Prefetch(
                name,
                pipe(
                    include_fields("pk"),
                    prepare_related_queryset,
                )(related_queryset),
                to_attr,
            )
        ),
    )


def prefetch_reverse_relationship(
    name, related_name, related_queryset, prepare_related_queryset=noop, to_attr=None
):
    """
    Efficiently prefetch a reverse relationship: one where the field on the "parent"
    queryset is not a concrete field - a foreign key from another object points at it.
    We need the `related_name` (ie the name of the relationship field on the related
    object) so that we can include this field in the query for the related objects,
    as Django will need it when it comes to stitch them together when the query
    is executed.
    """
    return pipe(
        include_fields("pk"),
        prefetch_related(
            Prefetch(
                name,
                pipe(
                    include_fields(related_name),
                    prepare_related_queryset,
                )(related_queryset),
                to_attr,
            )
        ),
    )


def prefetch_many_to_many_relationship(
    name, related_queryset, prepare_related_queryset=noop, to_attr=None
):
    """
    For many-to-many relationships, both sides of the relationship are non-concrete,
    so we don't need to do anything special with including fields. They are also
    symmetrical, so no need to differentiate between forward and reverse direction.
    """
    return pipe(
        include_fields("pk"),
        prefetch_related(
            Prefetch(
                name,
                pipe(
                    include_fields("pk"),
                    prepare_related_queryset,
                )(related_queryset),
                to_attr,
            )
        ),
    )


def auto_prefetch_relationship(name, prepare_related_queryset=noop, to_attr=None):
    """
    Given the name of a relationship, return a prepare function which introspects the
    relationship to discover its type and generates the correct set of
    `prefetch_related` and `include_fields` calls to apply to efficiently load it. A
    queryset function may also be passed, which will be applied to the related
    queryset.

    This is by far the most complicated part of the entire library. The reason
    it's so complicated is because Django's related object descriptors are
    inconsistent: each type has a slightly different way of accessing its related
    queryset, the name of the field on the other side of the relationship, etc.
    """

    def prepare(queryset):
        related_descriptor = getattr(queryset.model, name)

        if type(related_descriptor) in (
            ForwardOneToOneDescriptor,
            ForwardManyToOneDescriptor,
        ):
            return prefetch_forward_relationship(
                name,
                related_descriptor.field.related_model.objects.all(),
                prepare_related_queryset,
                to_attr,
            )(queryset)
        if type(related_descriptor) is ReverseOneToOneDescriptor:
            return prefetch_reverse_relationship(
                name,
                related_descriptor.related.field.name,
                related_descriptor.related.field.model.objects.all(),
                prepare_related_queryset,
                to_attr,
            )(queryset)
        if type(related_descriptor) is ReverseManyToOneDescriptor:
            return prefetch_reverse_relationship(
                name,
                related_descriptor.rel.field.name,
                related_descriptor.rel.field.model.objects.all(),
                prepare_related_queryset,
                to_attr,
            )(queryset)
        if type(related_descriptor) is ManyToManyDescriptor:
            field = related_descriptor.rel.field
            if related_descriptor.reverse:
                related_queryset = field.model.objects.all()
            else:
                related_queryset = field.target_field.model.objects.all()

            return prefetch_many_to_many_relationship(
                name,
                related_queryset,
                prepare_related_queryset,
                to_attr,
            )(queryset)

    return prepare



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
from setuptools import setup
import re
import os
import sys


name = "django-readers"
package = "django_readers"
description = "A lightweight function-oriented toolkit for better organisation of business logic and efficient selection and projection of data in Django projects."
url = "https://www.django-readers.org"
author = "DabApps"
author_email = "hello@dabapps.com"
license = "BSD"

with open("README.md") as f:
    readme = f.read()


def get_version(package):
    """
    Return package version as listed in `__version__` in `init.py`.
    """
    init_py = open(os.path.join(package, "__init__.py")).read()
    return re.search("^__version__ = ['\"]([^'\"]+)['\"]", init_py, re.MULTILINE).group(
        1
    )


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath
        for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [
        (dirpath.replace(package + os.sep, "", 1), filenames)
        for dirpath, dirnames, filenames in os.walk(package)
        if not os.path.exists(os.path.join(dirpath, "__init__.py"))
    ]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename) for filename in filenames])
    return {package: filepaths}


setup(
    name=name,
    version=get_version(package),
    url=url,
    license=license,
    description=description,
    long_description=readme,
    long_description_content_type="text/markdown",
    author=author,
    author_email=author_email,
    packages=get_packages(package),
    package_data=get_package_data(package),
    python_requires=">=3.6",
    install_requires=[
        "Django>=2.2",
    ],
    project_urls={
        "Changelog": "https://github.com/dabapps/django-readers/releases",
        "Issues": "https://github.com/dabapps/django-readers/issues",
    }
)



from django.core.exceptions import ObjectDoesNotExist
from django_readers.utils import map_or_apply, none_safe_attrgetter
from operator import attrgetter, methodcaller


def attr(name, *, transform_value=None, transform_value_if_none=False):
    def producer(instance):
        value = none_safe_attrgetter(name)(instance)
        if transform_value and (value is not None or transform_value_if_none):
            value = transform_value(value)
        return value

    return producer


method = methodcaller


def relationship(name, related_projector):
    """
    Given an attribute name and a projector, return a producer which plucks
    the attribute off the instance, figures out whether it represents a single
    object or an iterable/queryset of objects, and applies the given projector
    to the related object or objects.
    """

    def producer(instance):
        try:
            related = none_safe_attrgetter(name)(instance)
        except ObjectDoesNotExist:
            return None
        return map_or_apply(related, related_projector)

    return producer


def pk_list(name):
    """
    Given an attribute name (which should be a relationship field), return a
    producer which returns a list of the PK of each item in the relationship (or
    just a single PK if this is a to-one field, but this is an inefficient way of
    doing it).
    """
    return relationship(name, attrgetter("pk"))
